/* Pavel Surynek, 2008 */
/* bibox_main.cpp */

/*
   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least two free places. The code is not commented however it is simple enough to be understood
   without comments.

   The problem is given as an undirected graph, where each vertex of the graph is either free or
   occupied by a robot. A robot can move to a neighboring unoccupied vertex supposed that no other
   robot is simultaneously entering the target vertex. The task is to move robots preserving this
   condition to their final positions.

   This program is an implementation of the algorithm for solving the problem where the graph is
   biconnected and threre are at least two unoccupied vertices in the graph. The program runs in
   polynomial time with respect to the size of the input graph. More precisely it runs in O(|V|^3)
   where G=(V,E) is the input biconnected graph.
 */

#include "bibox.h"


int main(int UNUSED(argc), char **UNUSED(argv))
{
	clock_t begin, end;

	printf("Mulirobot %s\n", VERSION);
	printf("(c) 2008 %s\n\n", AUTHOR);

/*
	g.m_cycles.push_back(CYCLE());
	CYCLE *cycle = &g.m_cycles.back();

	cycle->m_x_id = 3; // Same 1
	cycle->m_y_id = 5; // Same 1 + 2

	g.add_vertex(1, 0, 6, 6);
	cycle->m_vertices.push_back(1);
	cycle->m_loop.push_back(1);
	g.add_vertex(2, 0, 5, 5);
	cycle->m_vertices.push_back(2);
	cycle->m_loop.push_back(2);
	g.add_vertex(3, 0, 1, 3);  // Same 1
	cycle->m_vertices.push_back(3);
	cycle->m_loop.push_back(3);
	g.add_vertex(4, 0, 2, 4);
	cycle->m_vertices.push_back(4);
	cycle->m_loop.push_back(4);
	g.add_vertex(5, 0, 8, 11);
	cycle->m_vertices.push_back(5);
	cycle->m_loop.push_back(5);
	g.add_vertex(6, 0, 7, 12);
	cycle->m_vertices.push_back(6);
	cycle->m_loop.push_back(6);
	g.add_vertex(7, 0, -1, 7);
	cycle->m_vertices.push_back(7);
	cycle->m_loop.push_back(7);
	g.add_vertex(8, 0, 0, 8);
	cycle->m_vertices.push_back(8);
	cycle->m_loop.push_back(8);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 4;
	cycle->m_y_id = 8;
	g.add_vertex(9, 1, 3, 1);
	cycle->m_vertices.push_back(9);
	cycle->m_loop.push_back(9);
	g.add_vertex(10, 1, 4, 2);
	cycle->m_vertices.push_back(10);
	cycle->m_loop.push_back(10);
	g.add_vertex(11, 1, 11, 0);
	cycle->m_vertices.push_back(11);
	cycle->m_loop.push_back(11);
	g.add_vertex(12, 1, 12, -1);
	cycle->m_vertices.push_back(12);
	cycle->m_loop.push_back(12);

	cycle->m_vertices.push_back(8);
	cycle->m_vertices.push_back(4);

	g.add_edge(1, 2, 0);
	g.add_edge(2, 3, 0);
	g.add_edge(3, 4, 0);
	g.add_edge(4, 5, 0);
	g.add_edge(5, 6, 0);
	g.add_edge(6, 7, 0);
	g.add_edge(7, 8, 0);
	g.add_edge(8, 1, 0);

	g.add_edge(4, 9, 1);
	g.add_edge(9, 10, 1);
	g.add_edge(10, 11, 1);
	g.add_edge(11, 12, 1);
	g.add_edge(12, 8, 1);
*/
/*
	g.m_cycles.push_back(CYCLE());
	CYCLE *cycle = &g.m_cycles.back();

	cycle->m_x_id = 2;
	cycle->m_y_id = 4; // Same 1

	g.add_vertex(1, 0, -1, 5);
	cycle->m_vertices.push_back(1);
	cycle->m_loop.push_back(1);
	g.add_vertex(2, 0, 0, 15);
	cycle->m_vertices.push_back(2);
	cycle->m_loop.push_back(2);
	g.add_vertex(3, 0, 1, 2);
	cycle->m_vertices.push_back(3);
	cycle->m_loop.push_back(3);
	g.add_vertex(4, 0, 2, 4); // Same 1
	cycle->m_vertices.push_back(4);
	cycle->m_loop.push_back(4);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 3;
	cycle->m_y_id = 4;
	g.add_vertex(5, 1, 3, 1);
	cycle->m_vertices.push_back(5);
	cycle->m_loop.push_back(5);
	g.add_vertex(6, 1, 4, 3);
	cycle->m_vertices.push_back(6);
	cycle->m_loop.push_back(6);

	cycle->m_vertices.push_back(4);
	cycle->m_vertices.push_back(3);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 5;
	cycle->m_y_id = 6;
	g.add_vertex(7, 2, 5, 6);
	cycle->m_vertices.push_back(7);
	cycle->m_loop.push_back(7);
	g.add_vertex(8, 2, 6, 13);
	cycle->m_vertices.push_back(8);
	cycle->m_loop.push_back(8);

	cycle->m_vertices.push_back(6);
	cycle->m_vertices.push_back(5);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 1;
	cycle->m_y_id = 8;
	g.add_vertex(9, 3, 15, 9);
	cycle->m_vertices.push_back(9);
	cycle->m_loop.push_back(9);
	g.add_vertex(10, 3, 14, 12);
	cycle->m_vertices.push_back(10);
	cycle->m_loop.push_back(10);
	g.add_vertex(11, 3, 13, 11);
	cycle->m_vertices.push_back(11);
	cycle->m_loop.push_back(11);
	g.add_vertex(12, 3, 12, 10);
	cycle->m_vertices.push_back(12);
	cycle->m_loop.push_back(12);
	g.add_vertex(13, 3, 11, 14);
	cycle->m_vertices.push_back(13);
	cycle->m_loop.push_back(13);
	g.add_vertex(14, 3, 10, 0);
	cycle->m_vertices.push_back(14);
	cycle->m_loop.push_back(14);
	g.add_vertex(15, 3, 9, -1);
	cycle->m_vertices.push_back(15);
	cycle->m_loop.push_back(15);

	cycle->m_vertices.push_back(8);
	cycle->m_vertices.push_back(6);
	cycle->m_vertices.push_back(4);
	cycle->m_vertices.push_back(1);

	g.add_edge(1, 2, 0);
	g.add_edge(2, 3, 0);
	g.add_edge(3, 4, 0);
	g.add_edge(4, 1, 0);
	g.add_edge(3, 5, 1);
	g.add_edge(5, 6, 1);
	g.add_edge(6, 4, 1);
	g.add_edge(5, 7, 2);
	g.add_edge(7, 8, 2);
	g.add_edge(8, 6, 2);
	g.add_edge(1, 9, 3);
	g.add_edge(9, 10, 3);
	g.add_edge(10, 11, 3);
	g.add_edge(11, 12, 3);
	g.add_edge(12, 13, 3);
	g.add_edge(13, 14, 3);
	g.add_edge(14, 15, 3);
	g.add_edge(15, 8, 3);
*/

	FILE *s;
	s = fopen("plan_statistics_lenght=3_free=90.txt", "w");

	FILE *b_lpg, *b_sgplan, *b_stan, *b_ipp, *b_max, *b_sat;
	b_lpg = fopen("batch_LPG_lenght=3_free=90.bat", "w");
	b_sgplan = fopen("batch_SGPLAN_lenght=3_free=90.bat", "w");
	b_stan = fopen("batch_STAN_lenght=3_free=90.bat", "w");
	b_ipp = fopen("batch_IPP_lenght=3_free=90.bat", "w");
	b_max = fopen("batch_MAX_lenght=3_free=90.bat", "w");
	b_sat = fopen("batch_SAT_lenght=3_free=90.bat", "w");

	int init = 4;
	int length = 4;
	int iter = 3;
	int free = 2;

//	for (iter = 10; iter > 0; --iter)
	{
//		for (length = 4; lenght <= 10; ++lenght)
		{
//			for (free = 2; free <= iter; ++free)
			{
				int cnt;
				for (cnt = 0; cnt < 10; ++cnt)
				{
					MOVE_COUNT = 0;
					STEP_COUNT = 0;
					GRAPH g, h;

					char name[1024];
					char filename[1024];
					
/*	
	g.print();

	solve_critical_free(&g);
*/
//	move_robot(4, 6, &g);
//	g.print();
/*
	move_free(1, 7, &g);
	g.print();	
*/

					generate_bipartite(init, iter, length, free, &g);
					free = 9 * g.m_vertices.size() / 10;
					if (free < 2)
					{
						free = 2;
					}
					generate_bipartite(init, iter, length, free, &h);

					if (h.m_vertices.size() > free)
					{
						sprintf(name, "multirobot_3_%d_90_c%d", iter, cnt);
						sprintf(filename, "%s.pddl", name);

						fprintf(b_sgplan, "./sgplan -o multirobot.pddl -f %s\n", filename);
						fprintf(b_lpg, "./lpg-td-1.0 -o multirobot.pddl -f %s -n 1\n", filename);
						fprintf(b_stan, "./stan multirobot.pddl %s\n", filename);
						fprintf(b_ipp, "./ipp -o multirobot.pddl -f %s\n", filename);
						fprintf(b_max, "./maxplan -domain multirobot.pddl -problem %s -cgdomain multirobot.pddl -cgproblem %s\n", filename, filename);
						fprintf(b_sat, "./satplan -domain multirobot.pddl -problem %s\n", filename);
						export_graph_2_PDDL(name, filename, &h);

						begin = clock();
						solve_critical_free(&h);
//	h.print();
						end = clock();

						printf("Problem parameters: %d %d %d %d\n", init, iter, length, free);
						printf("Size |V|=%lu |E|=%lu\n", h.m_vertices.size(), h.m_edges.size());
						printf("Move count: %d\n", MOVE_COUNT);
						printf("Step count: %d\n", STEP_COUNT);
						printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);

						fprintf(s, "%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%.3f\n", init, iter, length, free, h.m_vertices.size(), h.m_edges.size(), MOVE_COUNT, STEP_COUNT, (end - begin) / (double)CLOCKS_PER_SEC);
						fflush(s);
					}
					else
					{
						--cnt;
					}
				}
				fprintf(s, "\n");
				fprintf(b_lpg, "\n");
				fprintf(b_sgplan, "\n");
				fprintf(b_stan, "\n");
				fprintf(b_ipp, "\n");
				fprintf(b_max, "\n");
				fprintf(b_sat, "\n");
				fflush(s);
			}
		}
	}

	fclose(s);
	fclose(b_lpg);
	fclose(b_sgplan);
	fclose(b_stan);
	fclose(b_ipp);
	fclose(b_max);
	fclose(b_sat);

	return 0;
}
