/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* obox.cpp */

/*
   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least one free places. The main idea of the algorithm is to use precalculated sequences of moves for
   transposing robots and rotating robots along 3-cycles.
 */

#include "slider.h"
#include "tbox.h"
#include "obox.h"

//#define USE_XAOS	1

const double TIME_OUT = 1024.0;
int fnd, nfnd;

int DATABASE_HIT_COUNT;
int DATABASE_TOTAL_COUNT;
int COMP_SOLVED_COUNT;
int COMP_TOTAL_COUNT;

void extract_theta(int cycle_id, GRAPH *graph, IDS_list &A_loop, IDS_list &B_loop, IDS_list &C_loop)
{
	IDS_set dummy_fvertices;
	IDS_set dummy_fcycles;

	extract_theta(cycle_id, graph, A_loop, B_loop, C_loop, dummy_fvertices, dummy_fcycles);
}


void extract_theta(int cycle_id, GRAPH *graph, IDS_list &A_loop, IDS_list &B_loop, IDS_list &C_loop, IDS_set &fvertices, IDS_set &fcycles)
{
	CYCLE &cycle = graph->m_cycles[cycle_id];
	A_loop = cycle.m_loop;
	B_loop = cycle.m_bridge;
/*
	printf("A_loop:");
	print_IDS_list(A_loop);

	printf("B_loop:");
	print_IDS_list(B_loop);
*/
	fcycles.insert(cycle_id);

	IDS_list::const_iterator iv = cycle.m_bridge.begin();
	++iv;

	while (*iv != *cycle.m_bridge.rbegin())
	{
		fvertices.insert(*iv);
		++iv;
	}

	EDGES_set fedges;
	fedges.insert(EDGE(*B_loop.begin(), *B_loop.rbegin()));
/*
	printf("FEDGE:%d,%d\n", *B_loop.begin(), *B_loop.rbegin());
	printf("Ksi 1:%d,%d\n", cycle.m_x_id, cycle.m_y_id);
	printf("Fverices:");
	print_IDS_set(fvertices);
	printf("Fcycles:");
	print_IDS_set(fcycles);
*/
	bool connected = find_shortest_path(cycle.m_x_id, cycle.m_y_id, graph, fvertices, fcycles, fedges);

	if (connected)
	{
		int id = cycle.m_y_id;
		id = graph->m_vertices[id].m_prev_id;
	
		while (id != cycle.m_x_id)
		{
			C_loop.push_back(id);
			id = graph->m_vertices[id].m_prev_id;
		}
	}
	else
	{
		IDS_list::iterator f1 = B_loop.begin();
		IDS_list::iterator f2 = f1;
		++f2;

		fedges.insert(EDGE(*f1, *f2));

		IDS_list::reverse_iterator l1 = B_loop.rbegin();
		IDS_list::reverse_iterator l2 = l1;
		++l2;

		fedges.insert(EDGE(*l1, *l2));

		IDS_set separator;
		separator.insert(B_loop.begin(), B_loop.end());
		separator.erase(*B_loop.begin());
		separator.erase(*B_loop.rbegin());

		IDS_set dummy_fvertices;

		bool sep_con = find_shortest_path(cycle.m_x_id, separator, graph, dummy_fvertices, fcycles, fedges);

		int sp;

		for (IDS_set::const_iterator sep = separator.begin(); sep != separator.end(); ++sep)
		{
			if (graph->m_vertices[*sep].m_val != INT_MAX)
			{
				sp = *sep;
				break;
			}
		}

		IDS_list B_loop_alt1;

		for (IDS_list::const_iterator b = B_loop.begin(); b != B_loop.end(); ++b)
		{
			if (*b != sp)
			{
				B_loop_alt1.push_back(*b);
			}
			else
			{
				break;
			}
		}

		while (sp >= 0)
		{
			B_loop_alt1.push_back(sp);
			sp = graph->m_vertices[sp].m_prev_id;
		}

		IDS_list C_loop_alt1;

		sep_con = find_shortest_path(cycle.m_y_id, separator, graph, dummy_fvertices, fcycles, fedges);

		for (IDS_set::const_iterator sep = separator.begin(); sep != separator.end(); ++sep)
		{
			if (graph->m_vertices[*sep].m_val != INT_MAX)
			{
				sp = *sep;
				break;
			}
		}

		for (IDS_list::reverse_iterator b = B_loop.rbegin(); b != B_loop.rend(); ++b)
		{
			if (*b != sp)
			{
				C_loop_alt1.push_back(*b);
			}
			else
			{
				break;
			}
		}

		while (sp >= 0)
		{
			C_loop_alt1.push_back(sp);
			sp = graph->m_vertices[sp].m_prev_id;
		}

		C_loop_alt1.pop_back();
		C_loop_alt1.pop_front();

		C_loop_alt1.reverse();

		B_loop = B_loop_alt1;
		C_loop = C_loop_alt1;

//		assert(false);
	}
}


bool find_shortest_path(int source_id, int sink_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles, EDGES_set &fedges)
{
	IDS_set sink_ids;

	sink_ids.insert(sink_id);
	return find_shortest_path(source_id, sink_ids, graph, fvertices, fcycles, fedges);
}


bool find_shortest_path(int source_id, IDS_set &sink_ids, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles, EDGES_set &fedges)
{
	VERTICES_map::iterator v;

	for (v = graph->m_vertices.begin(); v != graph->m_vertices.end(); ++v)
	{
		v->second.m_val = INT_MAX;
		v->second.m_active = true;
	}

	VERTEX &source_vertex = graph->m_vertices[source_id];
	source_vertex.m_val = 0;
	source_vertex.m_prev_id = -1;

	VERTICES_set unfinished;
	unfinished.insert(source_id);

	while (!unfinished.empty())
	{
		VERTICES_set::iterator lowest, u;

		lowest = u = unfinished.begin();

		for (++u; u != unfinished.end(); ++u)
		{
			if (graph->m_vertices[*u].m_val < graph->m_vertices[*lowest].m_val)
			{
				lowest = u;
			}
		}

		int U, V;
		U = *lowest;

		VERTEX &lowest_vertex = graph->m_vertices[*lowest];
		int lowest_val = lowest_vertex.m_val;
		lowest_vertex.m_active = false;
		unfinished.erase(lowest);

		IDS_list &neighbors = lowest_vertex.m_neighbors;
		IDS_list::iterator n;

		for (n = neighbors.begin(); n != neighbors.end(); ++n)
		{
			VERTEX &n_vertex = graph->m_vertices[*n];
			V = *n;

			if (fvertices.find(n_vertex.m_id) == fvertices.end() && fcycles.find(n_vertex.m_cyc_id) == fcycles.end() && find_edge(U, V, fedges) == fedges.end())
			{
				if (n_vertex.m_active)
				{
					if (n_vertex.m_val > lowest_val + 1)
					{
						n_vertex.m_val = lowest_val + 1;
						n_vertex.m_prev_id = lowest_vertex.m_id;
						unfinished.insert(n_vertex.m_id);
					}
				}
			}
		}
		for (IDS_set::const_iterator sink_id = sink_ids.begin(); sink_id != sink_ids.end(); ++sink_id)
		{
			if (graph->m_vertices[*sink_id].m_val != INT_MAX)
			{
				return true;
			}
		}
	}

	return false;
}


EDGES_set::iterator find_edge(int U_id, int V_id, EDGES_set &edges)
{
	for (EDGES_set::iterator edge = edges.begin(); edge != edges.end(); ++edge)
	{
		if ((edge->m_U_id == U_id && edge->m_V_id == V_id) || (edge->m_U_id == V_id && edge->m_V_id == U_id))
		{
			return edge;
		}
	}

	return edges.end();
}


bool solve_slider_obox_theta(GRAPH *graph, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool use_theta, bool use_3_cycles)
{
	graph->GLOBAL_SOLUTION.clear();

	MOVE_COUNT = 0;
	STEP_COUNT = 0;

	DATABASE_HIT_COUNT = 0;
	DATABASE_TOTAL_COUNT = 0;
	COMP_SOLVED_COUNT = 0;
	COMP_TOTAL_COUNT = 0;

	IDS_set zero_nega_fvertices, zero_nega_fcycles;

	CYCLE *cycle_0 = &graph->m_cycles[0];

	IDS_list zero_path;

	VERTICES_map::const_iterator zero = find_final(0, graph);

	int first = graph->m_cycles[1].m_x_id;
	find_shortest_path(zero->second.m_id, first, graph, zero_nega_fvertices, zero_nega_fcycles);
	int z = first;//*cycle_0->m_vertices.begin();
	
	while (z >= 0)
	{
		zero_path.push_front(z);
		z = graph->m_vertices[z].m_prev_id;
	}

	IDS_list::iterator p1 = zero_path.begin();
	IDS_list::iterator p2 = p1;
	++p2;

	while (p2 != zero_path.end())
	{
		swap_finals(*p1, *p2, graph);
		++p1;
		++p2;
	}

	int c;

	IDS_set fcycles;
	int free = 0;

	for (c = graph->m_cycles.size() - 1; c >= 2; --c)
	{
		IDS_set theta_fvertices;
		IDS_list A_loop_save, B_loop_save, C_loop_save;
		IDS_list A_loop, B_loop, C_loop, avoid_loop;

		IDS_set fcycles_theta = fcycles;
		extract_theta(c, graph, A_loop_save, B_loop_save, C_loop_save, theta_fvertices, fcycles_theta);

		IDS_set theta_ids;

		theta_ids.insert(A_loop_save.begin(), A_loop_save.end());
		theta_ids.insert(B_loop_save.begin(), B_loop_save.end());
		theta_ids.insert(C_loop_save.begin(), C_loop_save.end());

		CYCLE *cycle = &graph->m_cycles[c];

		int unfin = -1;
		IDS_set dummy_fvertices;

		VERTICES_map::const_iterator f;
		f = find_robot(free, graph);
		move_free(f->second.m_id, cycle->m_x_id, graph, dummy_fvertices, fcycles);

		fcycles.insert(c);

		for (IDS_list::iterator l = cycle->m_loop.begin(); l != cycle->m_loop.end(); ++l)
		{
			if (graph->m_vertices[*l].m_robot != graph->m_vertices[*l].m_final)
			{
				unfin = *l;

				int robot = graph->m_vertices[unfin].m_final;
				bool inside_theta = false;

				for (IDS_set::const_iterator th = theta_ids.begin(); th != theta_ids.end(); ++th)
				{
					if (graph->m_vertices[*th].m_robot == robot)
					{
						inside_theta = true;
						break;
					}
				}

				int th_target = -1;
				
				if (inside_theta)
				{
					VERTICES_map::const_iterator r;
					r = find_robot(robot, graph);

					th_target = r->second.m_id;

					VERTICES_map::const_iterator f;
					f = find_robot(free, graph);
					
					IDS_set fvertices;
					fvertices.clear();
					fvertices.insert(th_target);

					move_free(f->second.m_id, cycle->m_x_id, graph, fvertices, fcycles);
				}
				else
				{
					EDGES_set dummy_fedges;

					VERTICES_map::const_iterator r = find_robot(robot, graph);

					IDS_set theta_ids_alowed = theta_ids;
					theta_ids_alowed.erase(cycle->m_x_id);

					IDS_set dummy_fvertices;
					find_shortest_path(r->second.m_id, theta_ids_alowed, graph, dummy_fvertices, fcycles, dummy_fedges);

					for (IDS_set::const_iterator th = theta_ids_alowed.begin(); th != theta_ids_alowed.end(); ++th)
					{
						if (graph->m_vertices[*th].m_val != INT_MAX)
						{
							th_target = *th;
							break;
						}
					}
					
					IDS_set theta_fvertices = theta_ids;

					theta_fvertices.erase(th_target);
					theta_fvertices.erase(cycle->m_x_id);

					move_robot(robot, th_target, 0, graph, theta_fvertices, fcycles);
					VERTICES_map::const_iterator f;
					f = find_robot(free, graph);

					IDS_set fvertices;
					fvertices.clear();
					fvertices.insert(th_target);
					move_free(f->second.m_id, cycle->m_x_id, graph, fvertices, fcycles);
				}
				A_loop = A_loop_save;
				B_loop = B_loop_save;
				C_loop = C_loop_save;

				IDS_list *pA_loop, *pB_loop, *pC_loop;

				if (A_loop.size() <= C_loop.size())
				{
					pA_loop = &C_loop;
					pB_loop = &B_loop;
					pC_loop = &A_loop;
				}
				else
				{
					pA_loop = &C_loop;
					pB_loop = &B_loop;
					pC_loop = &A_loop;
					
				}
				
				int A, B, C;

				A = pA_loop->size();
				B = pB_loop->size();
				C = pC_loop->size();

				IDS_vector theta_mapping;
				theta_mapping.push_back(*pB_loop->rbegin());
				pB_loop->pop_back();
				
				for (IDS_list::reverse_iterator v = pA_loop->rbegin(); v != pA_loop->rend(); ++v)
				{
					theta_mapping.push_back(*v);
				}
				for (IDS_list::iterator v = pB_loop->begin(); v != pB_loop->end(); ++v)
				{
					theta_mapping.push_back(*v);
				}
				for (IDS_list::reverse_iterator v = pC_loop->rbegin(); v != pC_loop->rend(); ++v)
				{
					theta_mapping.push_back(*v);
				}
/*				
				for (IDS_vector::const_iterator tm = theta_mapping.begin(); tm != theta_mapping.end(); ++tm)
				{
					printf("%d,", *tm);
				}
				printf("\n");
*/			
				int X = map_theta(th_target, theta_mapping);
				int Y = map_theta(unfin, theta_mapping);
				int Z = -1;
/*
				printf("ABC:%d,%d,%d\n", A, B, C);
				printf("XY:%d,%d\n", X, Y);
*/
				for (Z = 1; Z < A + B + C; ++Z)
				{
					if (Z != X && Z != Y)
					{
						bool alowed = true;
						
						for (IDS_list::iterator avoid = cycle->m_loop.begin(); avoid != cycle->m_loop.end(); ++avoid)
						{
							if (theta_mapping[Z] == *avoid)
							{
								alowed = false;
								break;
							}
						}
						
						if (alowed)
						{
							break;
						}
					}
				}
				
				SLIDER slider;
				init_slider(A, B, C, &slider);
/*
				printf("sABC:%d,%d,%d\n", slider.m_A, slider.m_B, slider.m_C);
				print_slider(&slider);
				printf("Before 1\n");
				print_slider(&slider);
				printf("XYZ:%d,%d,%d\n", theta_mapping[X], theta_mapping[Y], theta_mapping[Z]);
*/
				double timeout = TIME_OUT;
				
				bool in_database;
				FINE_SOLUTION sub_solution_theta, sub_solution;

//				solve_slider_tbox(&slider, database, sub_solution, in_database, timeout);
//				solve_slider_mit(&slider, sub_solution);

				if (use_theta)
				{
					if (use_3_cycles)
					{
						RECORDS_3_map::iterator record = database.m_3_records.find(RECORD_3KEY(slider.m_A, slider.m_B, slider.m_C, X, Y, Z));
						++DATABASE_TOTAL_COUNT;

						if (record != database.m_3_records.end())
						{
							++DATABASE_HIT_COUNT;
							sub_solution_theta = record->second;
						}
						else
						{
							SLIDER copy;
							copy_slider(&slider, &copy);

//							printf("XYZ:%d,%d,%d\n", X, Y, Z);
							bool sub_ok = solve_slider_3_tbox(&copy, X, Y, Z, database, sub_solution_theta, in_database, timeout);
							++COMP_TOTAL_COUNT;

							if (sub_ok)
							{
								++COMP_SOLVED_COUNT;
							}
							else
							{
								SLIDER copy2;
								copy_slider(&slider, &copy2);

								sub_solution_theta.clear();								
								
								if (!generate_3_cycle(&copy2, X, Y, Z, sub_solution_theta))
								{
									return false;
								}

								destroy_slider(&copy2);
							}

							destroy_slider(&copy);
						}
					}
					else
					{
						RECORDS_map::iterator record = database.m_records.find(RECORD_KEY(slider.m_A, slider.m_B, slider.m_C, X, Y));
						++DATABASE_TOTAL_COUNT;

						if (record != database.m_records.end())
						{
							++DATABASE_HIT_COUNT;
							sub_solution_theta = record->second;
						}
						else
						{
							SLIDER copy;
							copy_slider(&slider, &copy);

							swap_goal(X, Y, &copy);
//							printf("XY:%d,%d\n", X, Y);
/*
							print_slider(&slider);
							print_slider(&copy);
*/
#ifdef USE_XAOS
							bool sub_ok = solve_slider_tbox_xaos(&copy, database, sub_solution_theta, in_database, timeout);
#else							
							bool sub_ok = solve_slider_tbox_theta(&copy, database, sub_solution_theta, in_database, timeout);
#endif

							if (!sub_ok)
							{
								SLIDER copy2;
								copy_slider(&slider, &copy2);

								sub_solution_theta.clear();
								if (!generate_3_cycle(&copy2, X, Y, Z, sub_solution_theta))
								{
									return false;
								}
								destroy_slider(&copy2);
							}

							destroy_slider(&copy);
						}
					}
				}
				else
				{
					if (!generate_3_cycle(&slider, X, Y, Z, sub_solution_theta))
					{
						return false;
					}
				}

//				execute_fine_solution(&slider, sub_solution);
				translate_solution_theta(sub_solution_theta, theta_mapping, sub_solution);
				execute_fine_solution(graph, sub_solution);
				
				destroy_slider(&slider);
			}
		}
	}

//	return false;

	bool solved = solve_slider_obox_theta_last(graph, database, solution, fcycles, use_theta, use_3_cycles);

	if (!solved)
	{
		return false;
	}
	
	IDS_list::reverse_iterator rp1 = zero_path.rbegin();
	IDS_list::reverse_iterator rp2 = rp1;
	++rp2;

	while (rp2 != zero_path.rend())
	{
		swap_robots(*rp1, *rp2, graph);
		++rp1;
		++rp2;
	}

	solution = graph->GLOBAL_SOLUTION;

//	graph->print();

	return true;
}


bool solve_slider_obox_theta_last(GRAPH *graph, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, IDS_set &fcycles, bool use_theta, bool use_3_cycles)
{
	int free = 0;

	IDS_set dummy_fvertices;
	CYCLE *cycle = &graph->m_cycles[1];

	VERTICES_map::const_iterator f;
	f = find_robot(free, graph);
/*
	printf("Last X to free:%d\n", cycle->m_x_id);
	print_IDS_set(fcycles);
*/
	move_free(f->second.m_id, cycle->m_x_id, graph, dummy_fvertices, fcycles);

	IDS_list A_loop_last, B_loop_last, C_loop_last;
	IDS_set fcycles_theta = fcycles;
	IDS_set theta_fvertices;

	extract_theta(1, graph, A_loop_last, B_loop_last, C_loop_last, theta_fvertices, fcycles_theta);
/*
	printf("A_loop:");
	print_IDS_list(A_loop_last);
	
	printf("B_loop:");
	print_IDS_list(B_loop_last);

	printf("C_loop:");
	print_IDS_list(C_loop_last);
	printf("\n");
*/
	IDS_set theta_ids;

	theta_ids.insert(A_loop_last.begin(), A_loop_last.end());
	theta_ids.insert(B_loop_last.begin(), B_loop_last.end());
	theta_ids.insert(C_loop_last.begin(), C_loop_last.end());
	
	IDS_list A_loop = A_loop_last;
	IDS_list B_loop = B_loop_last;
	IDS_list C_loop = C_loop_last;

	IDS_list *pA_loop, *pB_loop, *pC_loop;

	if (A_loop.size() <= C_loop.size())
	{
		pA_loop = &C_loop;
		pB_loop = &B_loop;
		pC_loop = &A_loop;
	}
	else
	{
		pA_loop = &C_loop;
		pB_loop = &B_loop;
		pC_loop = &A_loop;
	}					
	
	int A, B, C;
	
	A = pA_loop->size();
	B = pB_loop->size();
	C = pC_loop->size();
	
	IDS_vector theta_mapping;
	theta_mapping.push_back(*pB_loop->rbegin());
	pB_loop->pop_back();
	
	for (IDS_list::reverse_iterator v = pA_loop->rbegin(); v != pA_loop->rend(); ++v)
	{
		theta_mapping.push_back(*v);
	}
	for (IDS_list::iterator v = pB_loop->begin(); v != pB_loop->end(); ++v)
	{
		theta_mapping.push_back(*v);
	}
	for (IDS_list::reverse_iterator v = pC_loop->rbegin(); v != pC_loop->rend(); ++v)
	{
		theta_mapping.push_back(*v);
	}
/*	
	printf("Last theta mapping: <-------------------------\n");
	for (IDS_vector::const_iterator tm = theta_mapping.begin(); tm != theta_mapping.end(); ++tm)
	{
		printf("%d,", *tm);
	}
	printf("\n");
*/
	SLIDER slider;
	init_slider(A, B, C, &slider);
	
	for (IDS_set::const_iterator th = theta_ids.begin(); th != theta_ids.end(); ++th)
	{
		int robot = graph->m_vertices[*th].m_robot;
		
		for (IDS_set::const_iterator th_target = theta_ids.begin(); th_target != theta_ids.end(); ++th_target)
		{
			if (robot == graph->m_vertices[*th_target].m_final)
			{
/*
				printf("%d ---> %d\n", *th, *th_target);
				printf("%d ---> %d\n", theta_mapping[*th], theta_mapping[*th_target]);
*/
				int k1, k2;
				for (k1 = 0; k1 < theta_mapping.size(); ++k1)
				{
					if (theta_mapping[k1] == *th)
					{
						break;
					}
				}
				for (k2 = 0; k2 < theta_mapping.size(); ++k2)
				{
					if (theta_mapping[k2] == *th_target)
					{
						break;
					}
				}
				slider.m_goal[k2] = k1;
			}
		}
	}

	double timeout = TIME_OUT;
				
	bool in_database;
	FINE_SOLUTION sub_solution_theta, sub_solution;

	bool solved;

	if (use_theta)
	{
		if (use_3_cycles)
		{
			SLIDER copy;
			copy_slider(&slider, &copy);

#ifdef USE_XAOS
			bool in_database;
			solved = solve_slider_3_tbox_xaos(&copy, database, sub_solution_theta, in_database, timeout);
#else
			solved = solve_slider_3_theta(&copy, database, sub_solution_theta, timeout);
#endif

			if (!solved)
			{
				sub_solution_theta.clear();

				SLIDER copy2;
				copy_slider(&slider, &copy2);

				solved = solve_slider_mit(&copy2, sub_solution_theta);

				if (!solved)
				{
					return false;
				}
				
				destroy_slider(&copy2);
			}

			destroy_slider(&copy);
		}
		else
		{

			SLIDER copy;
			copy_slider(&slider, &copy);

#ifdef USE_XAOS
			bool in_database;
			solved = solve_slider_tbox_xaos(&copy, database, sub_solution_theta, in_database, timeout);
#else
			solved = solve_slider_tbox_theta(&copy, database, sub_solution_theta, in_database, timeout);
#endif

			if (!solved)
			{
				sub_solution_theta.clear();

				SLIDER copy2;
				copy_slider(&slider, &copy2);

				solved = solve_slider_mit(&copy2, sub_solution_theta);

				if (!solved)
				{
					return false;
				}
				
				destroy_slider(&copy2);
			}

			destroy_slider(&copy);
		}
	}
	else
	{
		solved = solve_slider_mit(&slider, sub_solution_theta);

		if (!solved)
		{
			return false;
		}
	}

	execute_fine_solution(&slider, sub_solution_theta);

	translate_solution_theta(sub_solution_theta, theta_mapping, sub_solution);
	execute_fine_solution(graph, sub_solution);

	destroy_slider(&slider);

	return true;
}


int map_theta(int id, const IDS_vector &theta_mapping)
{
	for (unsigned int th = 0; th < theta_mapping.size(); ++th)
	{
		if (id == theta_mapping[th])
		{
			return th;
		}
	}

	return -1;
}


void translate_solution_theta(const FINE_SOLUTION &theta_solution, const IDS_vector &theta_mapping, FINE_SOLUTION &solution)
{
	for (FINE_SOLUTION::const_iterator th_move = theta_solution.begin(); th_move != theta_solution.end(); ++th_move)
	{
		assert(th_move->m_x < theta_mapping.size());
		assert(th_move->m_y < theta_mapping.size());

		FINE_MOVE move(theta_mapping[th_move->m_x], theta_mapping[th_move->m_y]);
		solution.push_back(move);
	}
}


void execute_fine_solution(GRAPH *graph, const FINE_SOLUTION &solution)
{
	for (FINE_SOLUTION::const_iterator move = solution.begin(); move != solution.end(); ++move)
	{
//		printf("%d --> %d\n", move->m_x, move->m_y);
		swap_robots(move->m_x, move->m_y, graph);
	}
}


void filter_fine_solution(GRAPH *graph, const FINE_SOLUTION &solution, FINE_SOLUTION &filtered_solution)
{
	for (FINE_SOLUTION::const_iterator move = solution.begin(); move != solution.end(); ++move)
	{
		if (swap_robots_empty(move->m_x, move->m_y, graph))
		{
			filtered_solution.push_back(*move);
		}
	}
}


void parallelize_fine_solution(FINE_SOLUTION &solution)
{
	FINE_SOLUTION::iterator m1 = solution.begin();

	for (++m1; m1 != solution.end(); ++m1)
	{
		for (FINE_SOLUTION::iterator m2 = solution.begin(); m2 != m1; ++m2)
		{
			if (FINE_MOVE::dependent(*m1, *m2))
			{
				if (m1->m_time < m2->m_time + 1)
				{
					m1->m_time = m2->m_time + 1;
				}
			}
		}
	}
}


bool solve_slider_3_theta(SLIDER *slider, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, double timeout)
{
	if (slider->m_Y <= 3)
	{
		printf("Too small first handle. Unable to solve. Terminating...\n");
		assert(false);
	}

	std::vector<int> forbid;

	SLIDER copy;
	copy_slider(slider, &copy);

	FINE_SOLUTION rotation_solution;

	int trans = calculate_goal_transpositions(&copy);

	if (trans % 2 == 1)  // odd permutation, we need to find odd cycle
	{
		if (copy.m_C1 % 2 == 1)
		{
			cycle_1_rotate_L(&copy, rotation_solution);
		}
		else
		{
			if (copy.m_C2 % 2 == 1)
			{
				cycle_2_rotate_L(&copy, rotation_solution);
			}
			else
			{
				if (copy.m_C3 % 2 == 1)
				{
					cycle_3_rotate_L(&copy, rotation_solution);
				}
				else
				{
					printf("Unable to solve, cannot obtain odd permutation using 3-cycles. Terminating...\n");
					assert(false);
				}
			}
		}
	}
	solution.insert(solution.end(), rotation_solution.begin(), rotation_solution.end());

	forbid.push_back(1);

	for (int i = 1; i < copy.m_V; ++i)
	{
//		printf("i:%d %d *** %d\n", i, copy.m_vrtxs[i], copy.m_goal[i]);
//		forbid.push_back(copy.m_vrtxs[i]);
		forbid.push_back(i);
      //	forbid.push_back(copy.m_goal[i]);
/*
		for (std::vector<int>::const_iterator f = forbid.begin(); f != forbid.end(); ++f)
		{
			printf("%d,", *f);
		}
		printf("\n");
*/
		if (copy.m_vrtxs[i] != copy.m_goal[i])
		{
			int z = -1;
			
			for (int j = 1; j < copy.m_V; ++j)
			{
				bool found = false;

				if (copy.m_vrtxs[j] == copy.m_goal[i])
				{
					found = true;
				}
				for (std::vector<int>::const_iterator f = forbid.begin(); f != forbid.end(); ++f)
				{
					if (*f == j)
					{
						found = true;
						break;
					}
				}
				if (!found)
				{
					z = j;
					break;
				}
			}

			if (z == -1)
			{
//				printf("Ending by z == -1 yyyyy 1\n");
				break;
			}

/*
			printf("*** (%d,%d,%d) ***\n", copy.m_goal[i], copy.m_vrtxs[i], z);
			printf("Lambda 1\n");
			print_slider(&copy);
			generate_3_cycle(&copy, copy.m_goal[i], copy.m_vrtxs[i], z, phase_solution);
			solution.insert(solution.end(), phase_solution.begin(), phase_solution.end());
			
			execute_fine_solution(&copy, phase_solution);
			printf("Lambda 2\n");
			print_slider(&copy);
*/
			int X, Y, Z;
			for (int k = 1; k < copy.m_V; ++k)
			{
				if (copy.m_goal[i] == copy.m_vrtxs[k])
				{
					X = k;
					break;
				}
			}
			Y = i;
			Z = z;
/*
			printf("XYZ:%d,%d,%d\n", X, Y, Z);
			printf("Bad:%d,%d,%d\n", copy.m_goal[i], copy.m_vrtxs[i], z);
*/
//			RECORDS_3_map::iterator record = database.m_3_records.find(RECORD_3KEY(copy.m_A, copy.m_B, copy.m_C, copy.m_goal[i], copy.m_vrtxs[i], z));
			RECORDS_3_map::iterator record = database.m_3_records.find(RECORD_3KEY(copy.m_A, copy.m_B, copy.m_C, X, Y, Z));
			++DATABASE_TOTAL_COUNT;

			if (record != database.m_3_records.end())
			{
				++DATABASE_HIT_COUNT;
				execute_fine_solution(&copy, record->second);
				solution.insert(solution.end(), record->second.begin(), record->second.end());
			}
			else
			{
				clock_t begin, end;

				SLIDER deep_copy;
				copy_slider(&copy, &deep_copy);

				FINE_SOLUTION fine_solution;

				bool in_database;

				begin = clock();
				bool solved = solve_slider_3_tbox(&deep_copy, X, Y, Z, database, fine_solution, in_database, timeout);
//				bool solved = solve_slider_3_tbox(&deep_copy, copy.m_goal[i], copy.m_vrtxs[i], z, database, fine_solution, in_database, timeout);
				++COMP_TOTAL_COUNT;

				end = clock();
				double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;
				timeout -= elapsed;
				
				if (solved)
				{
					++COMP_SOLVED_COUNT;
					in_database = false;

//					database.m_3_records[RECORD_3KEY(copy.m_A, copy.m_B, copy.m_C, copy.m_goal[i], copy.m_vrtxs[i], z)] = fine_solution;
					database.m_3_records[RECORD_3KEY(copy.m_A, copy.m_B, copy.m_C, X, Y, Z)] = fine_solution;
					solution.insert(solution.end(), fine_solution.begin(), fine_solution.end());
					execute_fine_solution(&copy, fine_solution);
				}
				else
				{
					destroy_slider(&copy);
					destroy_slider(&deep_copy);
					return false;
				}
								
				if (timeout < 0.0)
				{
					destroy_slider(&copy);
					destroy_slider(&deep_copy);
					return false;
				}

				destroy_slider(&deep_copy);
			}
/*
			printf("Lambda 2\n");
			print_slider(&copy);
*/
		}
	}

	destroy_slider(&copy);

	return true;
}


bool solve_slider_bibox_theta(GRAPH *graph, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool use_theta, bool use_3_cycles, bool mit)
{
	graph->GLOBAL_SOLUTION.clear();

	MOVE_COUNT = 0;
	STEP_COUNT = 0;

	DATABASE_HIT_COUNT = 0;
	DATABASE_TOTAL_COUNT = 0;
	COMP_SOLVED_COUNT = 0;
	COMP_TOTAL_COUNT = 0;

	IDS_set zero_nega_fvertices, zero_nega_fcycles;

	CYCLE *cycle_0 = &graph->m_cycles[0];

	IDS_list zero_path;
	IDS_list nega_path;

	VERTICES_map::const_iterator zero = find_final(0, graph);

	int first = graph->m_cycles[1].m_x_id;
//	printf("First 2:%d\n", first);

	find_shortest_path(zero->second.m_id, first, graph, zero_nega_fvertices, zero_nega_fcycles);
	int z = first;//*cycle_0->m_vertices.begin();
	
	while (z >= 0)
	{
		zero_path.push_front(z);
		z = graph->m_vertices[z].m_prev_id;
	}

	IDS_list::iterator p1 = zero_path.begin();
	IDS_list::iterator p2 = p1;
	++p2;

	while (p2 != zero_path.end())
	{
		swap_finals(*p1, *p2, graph);
		++p1;
		++p2;
	}

	zero_nega_fvertices.insert(first);

	if (!use_theta)
	{
		VERTICES_map::const_iterator nega = find_final(-1, graph);

		find_shortest_path(nega->second.m_id, *cycle_0->m_vertices.rbegin(), graph, zero_nega_fvertices, zero_nega_fcycles);
		z = *cycle_0->m_vertices.rbegin();
	
		while (z >= 0)
		{
			nega_path.push_front(z);
			z = graph->m_vertices[z].m_prev_id;
		}

		p1 = nega_path.begin();
		p2 = p1;
		++p2;

		while (p2 != nega_path.end())
		{
			swap_finals(*p1, *p2, graph);
			++p1;
			++p2;
		}
	}

	int c;

	IDS_set fcycles;
	int free = 0;

	for (c = graph->m_cycles.size() - 1; c >= 1; --c)
	{
		INDEPENDENT = true;
		CURRENT_CYCLE = c;

		IDS_set dummy_fvertices;

		CYCLE *cycle = &graph->m_cycles[c];

		int n_free_id = find_nearest_free(cycle->m_x_id, graph, dummy_fvertices, fcycles, cycle->m_loop);
		int n_free = graph->m_vertices[n_free_id].m_robot;

		IDS_list::reverse_iterator l;

		bool forbidden_free = false;

		for (l = cycle->m_loop.rbegin(); l != cycle->m_loop.rend(); ++l)
		{
			if (graph->m_vertices[*l].m_goal == n_free)
			{
				forbidden_free = true;
				break;
			}
		}

		if (forbidden_free)
		{
			n_free = free;
		}

		VERTICES_map::const_iterator ff = find_robot(n_free, graph);
		move_free(ff->second.m_id, cycle->m_x_id, graph, dummy_fvertices, fcycles);

		bool finished = true;

		for (l = cycle->m_loop.rbegin(); l != cycle->m_loop.rend(); ++l)
		{
			if (graph->m_vertices[*l].m_robot != graph->m_vertices[*l].m_final)
			{
				finished = false;
			}
		}

		if (!finished)
		{
			for (l = cycle->m_loop.rbegin(); l != cycle->m_loop.rend(); ++l)
			{
				int robot = graph->m_vertices[*l].m_final;
				VERTICES_map::const_iterator r = find_robot(robot, graph);

				if (r->second.m_cyc_id < c)
				{
					IDS_set fvertices;
					fcycles.insert(c);

					VERTICES_map::const_iterator r = find_robot(robot, graph);
					int n1_free_id = find_nearest_free(r->second.m_id, graph, dummy_fvertices, fcycles, cycle->m_loop);
					int n1_free = graph->m_vertices[n1_free_id].m_robot;
					move_robot(robot, cycle->m_x_id, n1_free, graph, fvertices, fcycles);
					fvertices.clear();
					fvertices.insert(cycle->m_x_id);
					int n2_free_id = find_nearest_free(cycle->m_y_id, graph, fvertices, fcycles, cycle->m_loop);
					int n2_free = graph->m_vertices[n2_free_id].m_robot;
					VERTICES_map::const_iterator f = find_robot(n2_free, graph);

					fvertices.clear();
					fvertices.insert(cycle->m_x_id);
					move_free(f->second.m_id, cycle->m_y_id, graph, fvertices, fcycles);
					rotate_cycle_CW(n2_free, cycle, graph);
				}
				else
				{
					IDS_set fvertices;

					fcycles.insert(c);

					int n3_free_id = find_nearest_free(cycle->m_x_id, graph, fvertices, fcycles, cycle->m_loop);
					int n3_free = graph->m_vertices[n3_free_id].m_robot;
					VERTICES_map::const_iterator f = find_robot(n3_free, graph);
					move_free(f->second.m_id, cycle->m_x_id, graph, fvertices, fcycles);
					
					int CWs = 0;

					while (graph->m_vertices[cycle->m_y_id].m_robot != robot)
					{
						rotate_cycle_CW(n3_free, cycle, graph);
						++CWs;
					}
					
					fvertices.clear();

					int out_id = find_nearest_outsider(cycle->m_y_id, cycle, graph, fvertices, fcycles);

					VERTICES_map::const_iterator r35 = find_robot(robot, graph);

					int n35_free_id = find_nearest_free(/*r35->second.m_id*/out_id, graph, fvertices, fcycles, cycle->m_loop);
					int n35_free = graph->m_vertices[n35_free_id].m_robot;

					move_robot(robot, out_id, n35_free, graph, fvertices, fcycles);

					fcycles.insert(c);
					fvertices.insert(out_id);

					int n4_free_id = find_nearest_free(cycle->m_x_id, graph, fvertices, fcycles, cycle->m_loop);
					int n4_free = graph->m_vertices[n4_free_id].m_robot;

					f = find_robot(n4_free, graph);
					move_free(f->second.m_id, cycle->m_x_id, graph, fvertices, fcycles);

					int i;

					for (i = 0; i < CWs; ++i)
					{
						rotate_cycle_CCW(n4_free, cycle, graph);
					}

					fvertices.clear();

					VERTICES_map::const_iterator r = find_robot(robot, graph);

					int n45_free_id = find_nearest_free(cycle->m_x_id/*r->second.m_id*/, graph, fvertices, fcycles, cycle->m_loop);
					int n45_free = graph->m_vertices[n45_free_id].m_robot;

					move_robot(robot, cycle->m_x_id, n45_free, graph, fvertices, fcycles);

					fvertices.clear();
					fvertices.insert(cycle->m_x_id);

					int n5_free_id = find_nearest_free(cycle->m_y_id, graph, fvertices, fcycles, cycle->m_loop);
					int n5_free = graph->m_vertices[n5_free_id].m_robot;

					f = find_robot(n5_free, graph);
					move_free(f->second.m_id, cycle->m_y_id, graph, fvertices, fcycles);

					rotate_cycle_CW(n5_free, cycle, graph);
				}
			}
		}

		if (INDEPENDENT)
		{
//			printf("Cycle: %d is independent !!!\n", CURRENT_CYCLE);
		}
		fcycles.insert(c);

		if (use_theta && c == 2)
		{
			bool solved;

			if (mit)
			{
				solved = solve_slider_obox_theta_last(graph, database, solution, fcycles, false, use_3_cycles);
			}
			else
			{
				solved = solve_slider_obox_theta_last(graph, database, solution, fcycles, use_theta, use_3_cycles);
			}

			if (!solved)
			{
				return false;
			}
			break;
		}
	}
/*
	printf("BIBOX fcycles:\n");
	print_IDS_set(fcycles);
*/
	if (!use_theta)
	{
		if (!solve_critical_last_free(graph, fcycles))
		{
			return false;
		}
	}

	IDS_list::reverse_iterator rp1 = nega_path.rbegin();
	IDS_list::reverse_iterator rp2 = rp1;
	++rp2;

	if (!use_theta)
	{
		while (rp2 != nega_path.rend())
		{
			swap_robots(*rp1, *rp2, graph);
			++rp1;
			++rp2;
		}
	}

	rp1 = zero_path.rbegin();
	rp2 = rp1;
	++rp2;

	while (rp2 != zero_path.rend())
	{
		swap_robots(*rp1, *rp2, graph);
		++rp1;
		++rp2;
	}

	solution = graph->GLOBAL_SOLUTION;
//	graph->print();

	return true;
}


bool solve_slider_tbox_theta(SLIDER *slider, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool &in_database, double timeout)
{
	SLIDER copy;

	copy_slider(slider, &copy);

	int p;

	in_database = true;

	for (int i = 1; i < copy.m_V; ++i)
	{
		if (copy.m_vrtxs[i] != copy.m_goal[i])
		{
			for (int j = 1; j < copy.m_V; j++)
			{
				if (copy.m_vrtxs[j] == copy.m_goal[i])
				{
					p = j;
					break;
				}
			}
//			printf("IP:%d,%d\n", i, p);
			RECORDS_map::iterator record = database.m_records.find(RECORD_KEY(slider->m_A, slider->m_B, slider->m_C, i, p));
			++DATABASE_TOTAL_COUNT;

			if (record != database.m_records.end())
			{
				++DATABASE_HIT_COUNT;
				execute_fine_solution(&copy, record->second);
				solution.insert(solution.end(), record->second.begin(), record->second.end());
			}
			else
			{
				if (timeout <= 0.0)
				{
					return false;
				}

				clock_t begin, end;

				SLIDER deep_copy;
				copy_slider(&copy, &deep_copy);

				for (int k = 0; k < deep_copy.m_V; ++k)
				{
					deep_copy.m_goal[k] = deep_copy.m_vrtxs[k];
				}
				swap_goal(i, p, &deep_copy);

				SOLUTION coarse_solution;
				FINE_SOLUTION fine_solution;
				init_solution(1024, &coarse_solution);

				begin = clock();
//				bool solved = solve_slider_ID_PL(&deep_copy, &coarse_solution, timeout);
				bool solved = solve_IDA_PL(&deep_copy, &coarse_solution, clock(), timeout);
				++COMP_TOTAL_COUNT;
				
				end = clock();
				double elapsed = (end - begin) / (double)CLOCKS_PER_SEC;
				timeout -= elapsed;
				
				if (solved)
				{
					++COMP_SOLVED_COUNT;
					in_database = false;

					solution2fine_solution(&copy, coarse_solution, fine_solution);
					database.m_records[RECORD_KEY(slider->m_A, slider->m_B, slider->m_C, i, p)] = fine_solution;
					solution.insert(solution.end(), fine_solution.begin(), fine_solution.end());
				}
				else
				{
					destroy_slider(&copy);
					destroy_slider(&deep_copy);
					return false;
				}
				
				destroy_solution(&coarse_solution);
				
				if (timeout < 0.0)
				{
					destroy_slider(&copy);
					destroy_slider(&deep_copy);
					return false;
				}

				destroy_slider(&deep_copy);
			}
		}
	}
	destroy_slider(&copy);

	return true;
}
