/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* obox.h */

/*
   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least one free places. The main idea of the algorithm is to use precalculated sequences of moves for
   transposing robots and rotating robots along 3-cycles.
 */

#ifndef __OBOX_H__
#define __OBOX_H__

#include <vector>

#include "bibox.h"
#include "tbox.h"

typedef std::vector<int> IDS_vector;

void print_IDS_list(const IDS_list &ids);
void print_IDS_set(const IDS_set &ids);

void extract_theta(int cycle_id, GRAPH *graph, IDS_list &A_loop, IDS_list &B_loop, IDS_list &C_loop);
void extract_theta(int cycle_id, GRAPH *graph, IDS_list &A_loop, IDS_list &B_loop, IDS_list &C_loop, IDS_set &fvertices, IDS_set &fcycles);
bool find_shortest_path(int source_id, int sink_id, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles, EDGES_set &fedges);
bool find_shortest_path(int source_id, IDS_set &sink_ids, GRAPH *graph, IDS_set &fvertices, IDS_set &fcycles, EDGES_set &fedges);
EDGES_set::iterator find_edge(int U_id, int V_id, EDGES_set &edges);
bool solve_slider_obox_theta(GRAPH *graph, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool use_theta, bool use_3_cycles);
int map_theta(int id, const IDS_vector &theta_mapping);
void translate_solution_theta(const FINE_SOLUTION &theta_solution, const IDS_vector &theta_mapping, FINE_SOLUTION &solution);

void execute_fine_solution(GRAPH *graph, const FINE_SOLUTION &solution);
void filter_fine_solution(GRAPH *graph, const FINE_SOLUTION &solution, FINE_SOLUTION &filtered_solution);
void parallelize_fine_solution(FINE_SOLUTION &solution);

bool solve_slider_3_theta(SLIDER *slider, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, double timeout);
bool solve_slider_obox_theta_last(GRAPH *graph, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, IDS_set &fcycles, bool use_theta, bool use_3_cycles);
bool solve_slider_bibox_theta(GRAPH *graph, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool use_theta, bool use_3_cycles, bool mit = false);
bool solve_slider_tbox_theta(SLIDER *slider, SOLUTION_DATABASE &database, FINE_SOLUTION &solution, bool &in_database, double timeout);

#endif
