/* Pavel Surynek, 2008 */
/* pavel.surynek@mff.cuni.cz */
/* obox_main.cpp */

/*
   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least one free places. The main idea of the algorithm is to use precalculated sequences of moves for
   transposing robots and rotating robots along 3-cycles.
 */

#include <list>

#include "slider.h"
#include "mit.h"
#include "bibox.h"
#include "obox.h"
#include "tbox.h"

SOLUTION_DATABASE database;

FINE_SOLUTION parallel_solution;

const int MULTIPLIER = 999;

void print_IDS_list(const IDS_list &ids)
{
	for (IDS_list::const_iterator id = ids.begin(); id != ids.end(); ++id)
	{
		printf("%d,", *id);
	}
	printf("\n");
}

void print_IDS_set(const IDS_set &ids)
{
	for (IDS_set::const_iterator id = ids.begin(); id != ids.end(); ++id)
	{
		printf("%d,", *id);
	}
	printf("\n");
}

int main(int UNUSED(argc), char **UNUSED(argv))
{
	clock_t begin, end;

	IDS_set exceptions;
/*
	exceptions.insert(919);
	exceptions.insert(370);
	exceptions.insert(281);
	exceptions.insert(857);
	exceptions.insert(124);
	exceptions.insert(60);
	exceptions.insert(368);
	exceptions.insert(739);
	exceptions.insert(619);
	exceptions.insert(31);
	exceptions.insert(927);
	exceptions.insert(367);
	exceptions.insert(788);
	exceptions.insert(584);
	exceptions.insert(226);
	exceptions.insert(280);

	exceptions.insert(505);
	exceptions.insert(586);

	exceptions.insert(886);
	exceptions.insert(915);
	exceptions.insert(335);
	exceptions.insert(690);
	exceptions.insert(926);
	exceptions.insert(540);
	exceptions.insert(782);
	exceptions.insert(58);
	exceptions.insert(167);
	exceptions.insert(456);
	exceptions.insert(42);
	exceptions.insert(421);
	exceptions.insert(784);
	exceptions.insert(315);
	exceptions.insert(956);
	exceptions.insert(170);
	exceptions.insert(996);
	exceptions.insert(305);
	exceptions.insert(327);
	exceptions.insert(336);
	exceptions.insert(545);
	exceptions.insert(814);
	exceptions.insert(808);
	exceptions.insert(178);
	exceptions.insert(932);
	exceptions.insert(795);
	exceptions.insert(570);
	exceptions.insert(378);
	exceptions.insert(601);
	exceptions.insert(317);
	exceptions.insert(756);
	exceptions.insert(441);
	exceptions.insert(444);
	exceptions.insert(729);
	exceptions.insert(771);
	exceptions.insert(481);
	exceptions.insert(675);


	exceptions.insert(286);
	exceptions.insert(117);
	exceptions.insert(709);
*/
/* exceptions for 9 */

	exceptions.insert(886);
	exceptions.insert(492);
	exceptions.insert(27);
	exceptions.insert(426);
	exceptions.insert(172);
	exceptions.insert(736);
	exceptions.insert(567);
	exceptions.insert(135);
	exceptions.insert(393);
	exceptions.insert(42);
	exceptions.insert(919);
	exceptions.insert(324);
	exceptions.insert(315);
	exceptions.insert(526);
	exceptions.insert(91);
	exceptions.insert(873);
	exceptions.insert(305);
	exceptions.insert(336);
	exceptions.insert(729);
	exceptions.insert(857);
	exceptions.insert(545);
	exceptions.insert(814);
	exceptions.insert(43);
	exceptions.insert(808);
	exceptions.insert(676);
	exceptions.insert(12);
	exceptions.insert(570);
	exceptions.insert(467);
	exceptions.insert(865);
	exceptions.insert(444);
	exceptions.insert(97);
	exceptions.insert(675);
	exceptions.insert(856);
	exceptions.insert(624);

/*
	srand(5787943);
	generate_bipartite(5, 8, 4, 1, &graph1);
*/
/*
	srand(1726745);
	generate_bipartite(5, 8, 4, 1, &graph1);
*/
/*
	srand(2325772);
	generate_bipartite(5, 10, 5, 1, &graph1);
*/
/*
	srand(4426427);
	generate_bipartite(5, 6, 4, 1, &graph1);
*/
/*
	srand(55263527);
	generate_bipartite(5, 8, 5, 2, &graph1);
*/
/*
	srand(72583527);
	generate_bipartite(6, 8, 6, 2, &graph1);
*/
/*
	srand(68439283);
	generate_bipartite(7, 8, 5, 2, &graph1);
*/
/*
	srand(38433283);
	generate_bipartite(9, 6, 5, 2, &graph1);
*/
/*
	srand(24433283);
	generate_bipartite(6, 6, 4, 2, &graph1);
*/
/*
	srand(24433283);
	generate_bipartite(6, 8, 4, 2, &graph1);
*/
	std::list<int> Seeds;

	for (int j = 0; j < 10000; ++j)
	{
		Seeds.push_back(random() % 1000);
	}

	int init = 9;
	int iter = 2;
	int loop = 7;
	int free = 2;

	int id = 0;

	char filename[1024];
//	sprintf(filename, "trend_theta_3-%d_%d.txt", init, iter);
//	sprintf(filename, "obox_theta_3_xaos-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "obox_theta_2-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "obox_theta_2_xaos-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "Results_bibox-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "Results_bibox_theta_2_xaos-%d_%d_%d.txt", init, iter, free);
	sprintf(filename, "Results_bibox_theta_2-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "Results_bibox_theta_3-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "Results_bibox_theta_3_xaos-%d_%d_%d.txt", init, iter, free);
//	sprintf(filename, "Results_bibox_theta_3_new-%d_%d_2.txt", init, iter);
//	sprintf(filename, "Results_bibox_theta_2_new-%d_%d_2.txt", init, iter);
//	sprintf(filename, "Results_bibox_theta_3-%d_%d_1.txt", init, iter);
//	sprintf(filename, "Results_bibox_theta_2_new-%d_%d_2.txt", init, iter);
//	sprintf(filename, "Results_bibox_theta_3_new-%d_%d_10p.txt", init, iter);
//	sprintf(filename, "Results_bibox_theta_2_new-%d_%d_10p.txt", init, iter);

//	sprintf(filename, "time_obox_mit-%d_%d_1.txt", init, iter);
//	sprintf(filename, "time_obox_theta_3-%d_%d_1.txt", init, iter);
//	sprintf(filename, "time_obox_theta_2-%d_%d_1.txt", init, iter);
//	sprintf(filename, "time_bibox_mit-%d_%d_1.txt", init, iter);
//	sprintf(filename, "time_bibox_theta_2-%d_%d_1.txt", init, iter);
//	sprintf(filename, "time_bibox_theta_3-%d_%d_1.txt", init, iter);

	FILE *f;
	f = fopen(filename, "a+");

	std::list<int>::iterator Seed;
	Seed = Seeds.begin();

	while (iter < 16)
//	while (free < 70)
	{
		for (int i = 0; i < 10; ++i)
		{
			int seed = *Seed;
			Seed++;
			printf("id:%d\n", id);
			printf("seed:%d\n", seed);
			
//			int seed = 126;

			if (exceptions.find(seed) == exceptions.end() && id > 96)
			{
				FINE_SOLUTION solution;
				FINE_SOLUTION filtered_solution;

				database.load("solution_database.txt");
				GRAPH graph1, graph2, graph1_copy;

				srand(seed);			       
				generate_bipartite(init, iter, loop, free, &graph1);

				srand(seed);			       
				generate_bipartite(init, iter, loop, free, &graph1_copy);

				begin = clock();
/*
				srand(seed);
				generate_bipartite(init, iter, loop, 1, &graph2);
*/
/*
				int free_perc = graph2.m_vertices.size() / 10;
				if (free_perc < 2)
				{
					free_perc = 2;
				}
				srand(seed);
*/
//				int free_perc;
				if (true) //free < graph2.m_vertices.size() && free >= 1)
				{
/*

					free_perc = free;

					printf("Free_perc:%d\n", free_perc);
*/
/*
					srand(seed);			       
					generate_bipartite(init, iter, loop, free_perc, &graph1);
*/
					bool solved;
// mit
/*
					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_obox_theta(&graph3, database, solution, false, false);
					}
*/
//					solved = solve_slider_obox_theta(&graph1, database, solution, false, false);

// theta_3
/*
					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_obox_theta(&graph3, database, solution, true, true);
				        }
*/
//					solved = solve_slider_obox_theta(&graph1, database, solution, true, true);

// theta_2
/*
					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_obox_theta(&graph3, database, solution, true, false);
					}
*/
//					solved = solve_slider_obox_theta(&graph1, database, solution, true, false);

// bibox
/*
					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_bibox_theta(&graph3, database, solution, false, false, false);
					}
*/
//					solved = solve_slider_bibox_theta(&graph1, database, solution, false, false, false);

// bibox_mit
/*
					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_bibox_theta(&graph3, database, solution, true, false, true);
					}
*/
//				        solved = solve_slider_bibox_theta(&graph1, database, solution, true, false, true);

// bibox_theta_3
/*

					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_bibox_theta(&graph3, database, solution, true, true, false);
					}
*/
//					solved = solve_slider_bibox_theta(&graph1, database, solution, true, true, false);


// bibox_theta_2
/*
					for (int m = 0; m < MULTIPLIER; ++m)
					{
						GRAPH graph3;

						srand(seed);
						generate_bipartite(init, iter, loop, free, &graph3);

						solved = solve_slider_bibox_theta(&graph3, database, solution, true, false, false);
					}
*/
					solved = solve_slider_bibox_theta(&graph1, database, solution, true, false, false);


					filter_fine_solution(&graph1_copy, solution, filtered_solution);
					parallelize_fine_solution(filtered_solution);

					printf("Solution size:%d\n", solution.size());
					printf("Filtered solution size:%d\n", filtered_solution.size());

					end = clock();

					if (solved)
					{
						database.save("solution_database.txt");
				
						fprintf(f, "%d\t%d\t%d\t%d\t%d\t%d\t", id, seed,  init, iter, loop, free);
						fprintf(f, "%d\t%d\t", graph1.m_vertices.size(), graph1.m_edges.size());
						fprintf(f, "%d\t", solution.size());
						fprintf(f, "%d\t%d\t", filtered_solution.size(), filtered_solution.rbegin()->m_time);
						fprintf(f, "%d\t%d\t", DATABASE_HIT_COUNT, DATABASE_TOTAL_COUNT);
						fprintf(f, "%d\t%d\t", COMP_SOLVED_COUNT, COMP_TOTAL_COUNT);
						fprintf(f, "%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);
					}
					else
					{
						fprintf(f, "%d\t%d\t-\t-\t-\t-\t", id, seed);
						fprintf(f, "-\t-\t");
						fprintf(f, "-\t");
						fprintf(f, "-\t-\t");
						fprintf(f, "-\t-\t");
						fprintf(f, "-\t-\t");
						fprintf(f, "-\n");
					}

				}
				else
				{

					fprintf(f, "%d\t%d\t-\t-\t-\t-\t", id, seed);
					fprintf(f, "-\t-\t");
					fprintf(f, "-\t");
					fprintf(f, "-\t-\t");
					fprintf(f, "-\t-\t");
					fprintf(f, "-\t-\t");
					fprintf(f, "-\n");

				}
			}
			else
			{
				fprintf(f, "%d\t%d\t-\t-\t-\t-\t", id, seed);
				fprintf(f, "-\t-\t");
				fprintf(f, "-\t");
				fprintf(f, "-\t-\t");
				fprintf(f, "-\t-\t");
				fprintf(f, "-\t-\t");
				fprintf(f, "-\n");

			}
			fflush(f);

			++id;
		}
		++iter;
//		++free;
	}

	fclose(f);

	return 0;
}
